@@ -1,7 +1,6 @@
-from __future__ import print_function, division, absolute_import
 import numpy as np
-from treetime import config as ttconf
-from treetime import MissingDataError
+from . import config as ttconf
+from . import MissingDataError
 from .treeanc import TreeAnc
 from .utils import numeric_date, DateConversion, datestring_from_numeric
 from .distribution import Distribution
@@ -23,8 +22,9 @@ class ClockTree(TreeAnc):
     is converted to the most likely time of the internal nodes.
     """
 
-    def __init__(self,  dates=None, debug=False, real_dates=True, precision='auto',
-                 branch_length_mode='joint', use_covariation=False, *args, **kwargs):
+    def __init__(self, *args, dates=None, debug=False, real_dates=True, precision_fft = 'auto',
+                precision='auto', precision_branch='auto', branch_length_mode='joint', use_covariation=False,
+                use_fft=True,**kwargs):
 
         """
         ClockTree constructor
@@ -52,6 +52,13 @@ class ClockTree(TreeAnc):
             When not specified, this will default to 1 for short sequences and 2
             for long sequences with L>1e4
 
+        precision_fft : int
+            When calculating convolutions using the FFT approach a regular
+            discrete grid needs to be chosen. To optimize the calculation the size is not
+            set to a fixed number but is determined by the FWHM of the distributions.
+            The number of points desired to span the width of the FWHM of a distribution
+            can be specified explicitly by precision_fft (default is 200).
+
          branch_length_mode : str
             determines whether branch length are calculated using the 'joint' ML,
             'marginal' ML, or branch length of the input tree ('input').
@@ -60,8 +67,12 @@ class ClockTree(TreeAnc):
             determines whether root-to-tip regression accounts for covariance
             introduced by shared ancestry.
 
+        use_fft: boolean
+            Use FFT for calculation of convolution integrals if true (default).
+            The alternative is kept to be able to reproduce previous behavior.
+
          **kwargs:
-            Key word argments needed to construct parent class (TreeAnc)
+            Key word arguments passed on to the parent class (TreeAnc)
 
         """
         super(ClockTree, self).__init__(*args, **kwargs)
@@ -71,6 +82,7 @@ class ClockTree(TreeAnc):
         self.debug=debug
         self.real_dates = real_dates
         self.date_dict = dates
+        self.use_fft = use_fft
         self._date2dist = None  # we do not know anything about the conversion
         self.tip_slack = ttconf.OVER_DISPERSION  # extra number of mutations added
                                                  # to terminal branches in covariance calculation
@@ -80,6 +92,7 @@ class ClockTree(TreeAnc):
         self.clock_model=None
         self.use_covariation=use_covariation # if false, covariation will be ignored in rate estimates.
         self._set_precision(precision)
+        self._set_precision_fft(precision_fft, precision_branch)
         self._assign_dates()
 
 
@@ -135,7 +148,7 @@ class ClockTree(TreeAnc):
 
     def _set_precision(self, precision):
         '''
-        function that sets precision to an (hopfully) reasonable guess based
+        function that sets precision to a (hopefully) reasonable guess based
         on the length of the sequence if not explicitly set
         '''
         # if precision is explicitly specified, use it.
@@ -177,6 +190,28 @@ class ClockTree(TreeAnc):
             self.branch_grid_points = ttconf.BRANCH_GRID_SIZE
             self.n_integral = ttconf.N_INTEGRAL
 
+    def _set_precision_fft(self, precision_fft, precision_branch='auto'):
+            '''
+            function to set the number of grid points for the minimal FWHM window and branch grid
+            when calculating the marginal distribution using the FFT-based approach
+            The default parameters ttconf.FFT_FWHM_GRID_SIZE and branch_grid_points determined in set_precision
+            are used unless an integer value is specified using precision_fft and precision_branch
+            '''
+
+            if type(precision_fft) is int:
+                self.logger("ClockTree.init._set_precision_fft: setting fft grid size explicitly,"
+                        " fft_grid_points=%.3e"%(precision_fft), 2)
+                self.fft_grid_size = precision_fft
+            elif precision_fft!='auto':
+                raise ValueError(f"ClockTree: precision_fft needs to be either 'auto' or an integer, got '{precision_fft}'.")
+            else:
+                self.fft_grid_size = ttconf.FFT_FWHM_GRID_SIZE
+            if type(precision_branch) is int:
+                self.logger("ClockTree.init._set_precision_fft: setting branch grid size explicitly,"
+                        " branch_grid_points=%.3e"%(precision_branch), 2)
+                self.branch_grid_points = precision_branch
+
+
     @property
     def date2dist(self):
         return self._date2dist
@@ -268,7 +303,7 @@ class ClockTree(TreeAnc):
 
         # set the None  for the date-related attributes in the internal nodes.
         # make interpolation objects for the branches
-        self.logger('ClockTree.init_date_constraints: Initializing branch length interpolation objects...',3)
+        self.logger('ClockTree.init_date_constraints: Initializing branch length interpolation objects...',2)
         has_clock_length = []
         for node in self.tree.find_clades(order='postorder'):
             if node.up is None:
@@ -278,10 +313,8 @@ class ClockTree(TreeAnc):
                 # copy the merger rate and gamma if they are set
                 if hasattr(node,'branch_length_interpolator') and node.branch_length_interpolator is not None:
                     gamma = node.branch_length_interpolator.gamma
-                    merger_cost = node.branch_length_interpolator.merger_cost
                 else:
                     gamma = 1.0
-                    merger_cost = None
 
                 if self.branch_length_mode=='marginal':
                     node.profile_pair = self.marginal_branch_profile(node)
@@ -289,16 +322,17 @@ class ClockTree(TreeAnc):
                     self.add_branch_state(node)
 
                 node.branch_length_interpolator = BranchLenInterpolator(node, self.gtr,
-                            pattern_multiplicity = self.data.multiplicity, min_width=self.min_width,
-                            one_mutation=self.one_mutation, branch_length_mode=self.branch_length_mode)
+                            pattern_multiplicity = self.data.multiplicity(mask=node.mask), min_width=self.min_width,
+                            one_mutation=self.one_mutation, branch_length_mode=self.branch_length_mode,
+                            n_grid_points = self.branch_grid_points)
 
-                node.branch_length_interpolator.merger_cost = merger_cost
                 node.branch_length_interpolator.gamma = gamma
 
         # use covariance in clock model only after initial timetree estimation is done
         use_cov = (np.sum(has_clock_length) > len(has_clock_length)*0.7) and self.use_covariation
         self.get_clock_model(covariation=use_cov, slope=clock_rate)
 
+        self.logger('ClockTree.init_date_constraints: node date constraints objects...', 2)
         # make node distribution objects
         for node in self.tree.find_clades(order="postorder"):
             # node is constrained
@@ -320,7 +354,7 @@ class ClockTree(TreeAnc):
                 node.date_constraint = None
 
 
-    def make_time_tree(self, time_marginal=False, clock_rate=None, **kwargs):
+    def make_time_tree(self, time_marginal=False, clock_rate=None, assign_dates=True, **kwargs):
         '''
         Use the date constraints to calculate the most likely positions of
         unconstrained nodes.
@@ -340,11 +374,14 @@ class ClockTree(TreeAnc):
         self.init_date_constraints(clock_rate=clock_rate, **kwargs)
 
         if time_marginal:
-            self._ml_t_marginal(assign_dates = time_marginal=="assign")
+            self._ml_t_marginal(assign_dates = assign_dates)
         else:
             self._ml_t_joint()
 
-        self.convert_dates()
+        self.tree.positional_LH = self.timetree_likelihood(time_marginal)
+
+        if assign_dates or not time_marginal:
+            self.convert_dates()
 
 
     def _ml_t_joint(self):
@@ -390,15 +427,28 @@ class ClockTree(TreeAnc):
                     # Cx.y is the branch length corresponding the optimal subtree
                     bl = node.branch_length_interpolator.x
                     x = bl + node.date_constraint.peak_pos
-                    node.joint_pos_Lx = Distribution(x, node.branch_length_interpolator(bl),
-                                                     min_width=self.min_width, is_log=True)
+                    if hasattr(self, 'merger_model') and self.merger_model:
+                        node.joint_pos_Lx =  Distribution(x, -self.merger_model.integral_merger_rate(node.date_constraint.peak_pos)
+                                                + node.branch_length_interpolator(bl), min_width=self.min_width, is_log=True)
+                    else:
+                        node.joint_pos_Lx =  Distribution(x, node.branch_length_interpolator(bl), min_width=self.min_width, is_log=True)
                     node.joint_pos_Cx = Distribution(x, bl, min_width=self.min_width) # map back to the branch length
                 else: # all nodes without precise constraint but positional information
                     msgs_to_multiply = [node.date_constraint] if node.date_constraint is not None else []
-                    msgs_to_multiply.extend([child.joint_pos_Lx for child in node.clades
-                                             if child.joint_pos_Lx is not None])
+                    child_messages = [child.joint_pos_Lx for child in node.clades if child.joint_pos_Lx is not None]
+                    msgs_to_multiply.extend(child_messages)
+                    ## When a coalescent model is being used, the cost of having no merger events along the branch
+                    ## and one at the node at time t is also factored in: -np.log((gamma(t) * np.exp**-I(t))**(k-1)),
+                    ## where k is the number of branches that merge at node t, gamma(t) is the total_merger_rate
+                    ## at time t and I(t) is the integral of the merger_rate (rate of a given lineage converging)
+                    ## evaluated at position t. (Note that the integral of the merger rate is in fact calculated
+                    ## for k branches, but due to the fact that inner branches overlap at time t one can be removed
+                    ## resulting in the exponent (k-1))
+                    if hasattr(self, 'merger_model') and self.merger_model:
+                        time_points = np.unique(np.concatenate([msg.x for msg in msgs_to_multiply]))
+                        msgs_to_multiply.append(self.merger_model.node_contribution(node, time_points))
 
-                    # subtree likelihood given the node's constraint and child messages
+                    # msgs_to_multiply combined returns the subtree likelihood given the node's constraint and child messages
                     if len(msgs_to_multiply) == 0: # there are no constraints
                         node.joint_pos_Lx = None
                         node.joint_pos_Cx = None
@@ -409,6 +459,10 @@ class ClockTree(TreeAnc):
                         subtree_distribution = msgs_to_multiply[0]
 
                     if node.up is None: # this is the root, set dates
+                        if hasattr(self, 'merger_model') and self.merger_model:
+                            # Removed merger rate must be added back at the root as nolonger an internal node
+                            subtree_distribution = Distribution.multiply([subtree_distribution, Distribution(subtree_distribution.x,
+                                                    self.merger_model.integral_merger_rate(subtree_distribution.x), is_log=True)])
                         subtree_distribution._adjust_grid(rel_tol=self.rel_tol_prune)
                         # set root position and joint likelihood of the tree
                         node.time_before_present = subtree_distribution.peak_pos
@@ -429,6 +483,18 @@ class ClockTree(TreeAnc):
                         node.joint_pos_Lx = res
                         node.joint_pos_Cx = res_t
 
+                # construct the inverse cumulant distribution for the branch number estimates
+                from scipy.interpolate import interp1d
+                if node.date_constraint is not None and node.date_constraint.is_delta:
+                    node.joint_inverse_cdf=interp1d([0,1], node.date_constraint.peak_pos*np.ones(2), kind="linear")
+                elif isinstance(subtree_distribution, Distribution):
+                    dt = np.diff(subtree_distribution.x)
+                    y = subtree_distribution.prob_relative(subtree_distribution.x)
+                    int_y = np.concatenate(([0], np.cumsum(dt*(y[1:]+y[:-1])/2.0)))
+                    int_y/=int_y[-1]
+                    node.joint_inverse_cdf = interp1d(int_y, subtree_distribution.x, kind="linear")
+                    #node.joint_cdf = interp1d(subtree_distribution.x, int_y, kind="linear")
+
 
         # go through the nodes from root towards the leaves and assign joint ML positions:
         self.logger("ClockTree - Joint reconstruction:  Propagating root -> leaves...", 2)
@@ -438,46 +504,56 @@ class ClockTree(TreeAnc):
                 continue # the position was already set on the previous step
 
             if node.joint_pos_Cx is None: # no constraints or branch is bad - reconstruct from the branch len interpolator
-                node.branch_length = node.branch_length_interpolator.peak_pos
+                if hasattr(self, 'merger_model') and self.merger_model and node.up is not None:
+                    ##add merger_cost if using the coalescent model
+                    merger_cost = Distribution(node.branch_length_interpolator.x, self.merger_model.cost(node.time_before_present,
+                                    node.branch_length_interpolator.x, multiplicity=len(node.up.clades)), is_log=True)
+                    node.branch_length = Distribution.multiply([merger_cost, node.branch_length_interpolator]).peak_pos
+                else:
+                    node.branch_length = node.branch_length_interpolator.peak_pos
             elif node.date_constraint is not None and node.date_constraint.is_delta:
                 node.branch_length = node.up.time_before_present - node.date_constraint.peak_pos
             elif isinstance(node.joint_pos_Cx, Distribution):
                 # NOTE the Lx distribution is the likelihood, given the position of the parent
                 # (Lx.x = parent position, Lx.y = LH of the node_pos given Lx.x,
                 # the length of the branch corresponding to the most likely
-                # subtree is node.Cx(node.time_before_present))
+                # subtree is node.Cx(node.up.time_before_present))
                 # subtree_LH = node.joint_pos_Lx(node.up.time_before_present)
                 node.branch_length = node.joint_pos_Cx(max(node.joint_pos_Cx.xmin,
-                                            node.up.time_before_present)+ttconf.TINY_NUMBER)
+                                                           node.up.time_before_present))
+
+            # clean up tiny negative branch length, warn against bigger ones.
+            if node.branch_length<0:
+                if node.branch_length>-2*ttconf.TINY_NUMBER:
+                    self.logger(f"ClockTree - Joint reconstruction: correcting rounding error of {node.name} bl={node.branch_length:1.2e}", 4)
+                else:
+                    self.logger(f"ClockTree - Joint reconstruction: NEGATIVE BRANCH LENGTH {node.name} bl={node.branch_length:1.2e}", 2, warn=True)
+                node.branch_length = 0
 
             node.time_before_present = node.up.time_before_present - node.branch_length
             node.clock_length = node.branch_length
 
-            # just sanity check, should never happen:
-            if node.branch_length < 0 or node.time_before_present < 0:
-                if node.branch_length<0 and node.branch_length>-ttconf.TINY_NUMBER:
-                    self.logger("ClockTree - Joint reconstruction: correcting rounding error of %s"%node.name, 4)
-                    node.branch_length = 0
-
-        self.tree.positional_joint_LH = self.timetree_likelihood()
         # cleanup, if required
         if not self.debug:
             _cleanup()
 
 
-    def timetree_likelihood(self):
+    def timetree_likelihood(self, time_marginal):
         '''
         Return the likelihood of the data given the current branch length in the tree
         '''
-        LH = 0
-        for node in self.tree.find_clades(order='preorder'):  # sum the likelihood contributions of all branches
-            if node.up is None: # root node
-                continue
-            LH -= node.branch_length_interpolator(node.branch_length)
+        if time_marginal:
+            LH =  -self.tree.root.marginal_pos_LH.peak_val
+        else:
+            LH = 0
+            for node in self.tree.find_clades(order='preorder'):  # sum the likelihood contributions of all branches
+                if node.up is None: # root node
+                    continue
+                LH -= node.branch_length_interpolator(node.branch_length)
 
-        # add the root sequence LH and return
-        if self.aln:
-            LH += self.gtr.sequence_logLH(self.tree.root.cseq, pattern_multiplicity=self.data.multiplicity)
+            # add the root sequence LH and return
+            if self.aln and self.sequence_reconstruction:
+                LH += self.gtr.sequence_logLH(self.tree.root.cseq, pattern_multiplicity=self.data.multiplicity())
         return LH
 
 
@@ -508,7 +584,6 @@ class ClockTree(TreeAnc):
             root.
 
         """
-
         def _cleanup():
             for node in self.tree.find_clades():
                 try:
@@ -519,7 +594,7 @@ class ClockTree(TreeAnc):
                 except:
                     pass
 
-
+        method = 'FFT' if self.use_fft else 'explicit'
         self.logger("ClockTree - Marginal reconstruction:  Propagating leaves -> root...", 2)
         # go through the nodes from leaves towards the root:
         for node in self.tree.find_clades(order='postorder'):  # children first, msg to parents
@@ -535,14 +610,27 @@ class ClockTree(TreeAnc):
                     node.subtree_distribution = node.date_constraint
                     bl = node.branch_length_interpolator.x
                     x = bl + node.date_constraint.peak_pos
-                    node.marginal_pos_Lx = Distribution(x, node.branch_length_interpolator(bl),
-                                                        min_width=self.min_width, is_log=True)
-
+                    if hasattr(self, 'merger_model') and self.merger_model:
+                        node.marginal_pos_Lx =  Distribution(x, -self.merger_model.integral_merger_rate(node.date_constraint.peak_pos)
+                                                    +node.branch_length_interpolator(bl), min_width=self.min_width, is_log=True)
+                    else:
+                        node.marginal_pos_Lx =  Distribution(x, node.branch_length_interpolator(bl), min_width=self.min_width, is_log=True)
                 else: # all nodes without precise constraint but positional information
                       # subtree likelihood given the node's constraint and child msg:
                     msgs_to_multiply = [node.date_constraint] if node.date_constraint is not None else []
                     msgs_to_multiply.extend([child.marginal_pos_Lx for child in node.clades
                                              if child.marginal_pos_Lx is not None])
+                    ## When a coalescent model is being used, the cost of having no merger events along the branch
+                    ## and one at the node at time t is also factored in: -np.log((gamma(t) * np.exp**-I(t))**(k-1)),
+                    ## where k is the number of branches that merge at node t, gamma(t) is the total_merger_rate
+                    ## at time t and I(t) is the integral of the merger_rate (rate of a given lineage converging)
+                    ## evaluated at position t. (Note that the integral of the merger rate is in fact calculated
+                    ## for k branches, but due to the fact that inner branches overlap at time t one can be removed
+                    ## resulting in the exponent (k-1))
+                    if hasattr(self, 'merger_model') and self.merger_model:
+                        time_points = np.unique(np.concatenate([msg.x for msg in msgs_to_multiply]))
+                        # set multiplicity of node to number of good child branches
+                        msgs_to_multiply.append(self.merger_model.node_contribution(node, time_points))
 
                     # combine the different msgs and constraints
                     if len(msgs_to_multiply)==0:
@@ -557,53 +645,83 @@ class ClockTree(TreeAnc):
                     if node.up is None: # this is the root, set dates
                         node.subtree_distribution._adjust_grid(rel_tol=self.rel_tol_prune)
                         node.marginal_pos_Lx = node.subtree_distribution
-                        node.marginal_pos_LH = node.subtree_distribution
-                        self.tree.positional_marginal_LH = -node.subtree_distribution.peak_val
+                        if hasattr(self, 'merger_model') and self.merger_model:
+                            # Removed merger rate must be added back at the root as nolonger an internal node
+                            node.marginal_pos_LH = Distribution.multiply([node.subtree_distribution, Distribution(node.subtree_distribution.x,
+                                                    self.merger_model.integral_merger_rate(node.subtree_distribution.x), is_log=True)])
+                        else:
+                            node.marginal_pos_LH = node.subtree_distribution
                     else: # otherwise propagate to parent
-                        res, res_t = NodeInterpolator.convolve(node.subtree_distribution,
+                        if self.use_fft:
+                            res, res_t = NodeInterpolator.convolve_fft(node.subtree_distribution,
+                                        node.branch_length_interpolator, self.fft_grid_size), None
+                        else:
+                            res, res_t = NodeInterpolator.convolve(node.subtree_distribution,
                                         node.branch_length_interpolator,
                                         max_or_integral='integral',
                                         n_grid_points = self.node_grid_points,
                                         n_integral=self.n_integral,
                                         rel_tol=self.rel_tol_refine)
-                        res._adjust_grid(rel_tol=self.rel_tol_prune)
+                            res._adjust_grid(rel_tol=self.rel_tol_prune)
                         node.marginal_pos_Lx = res
 
         self.logger("ClockTree - Marginal reconstruction:  Propagating root -> leaves...", 2)
         from scipy.interpolate import interp1d
         for node in self.tree.find_clades(order='preorder'):
 
-            ## The root node
-            if node.up is None:
+            ## If a delta constraint in known no further work required
+            if (node.date_constraint is not None) and (not node.bad_branch) and node.date_constraint.is_delta:
+                node.marginal_pos_LH = node.date_constraint
+                node.msg_from_parent = None #if internal node has a delta constraint noprevious information passed on
+            elif node.up is None:
                 node.msg_from_parent = None # nothing beyond the root
             # all other cases (All internal nodes + unconstrained terminals)
-            elif node.date_constraint is not None and node.date_constraint.is_delta:
-                node.marginal_pos_LH = node.date_constraint
             else:
                 parent = node.up
-                # messages from the complementary subtree (iterate over all sister nodes)
-                complementary_msgs = [sister.marginal_pos_Lx for sister in parent.clades
-                                            if (sister != node) and (sister.marginal_pos_Lx is not None)]
 
-                # if parent itself got smth from the root node, include it
-                if parent.msg_from_parent is not None:
-                    complementary_msgs.append(parent.msg_from_parent)
+                msg_parent_to_node =None
+                if node.marginal_pos_Lx is not None:
+                    # messages from the complementary subtree (iterate over all sister nodes)
+                    complementary_msgs = [parent.date_constraint] if parent.date_constraint is not None else []
+                    complementary_msgs.extend([sister.marginal_pos_Lx for sister in parent.clades
+                                                if (sister != node) and (sister.marginal_pos_Lx is not None)])
 
-                if len(complementary_msgs):
-                    msg_parent_to_node = NodeInterpolator.multiply(complementary_msgs)
-                    msg_parent_to_node._adjust_grid(rel_tol=self.rel_tol_prune)
-                else:
+                    # if parent itself got smth from the root node, include it
+                    if parent.msg_from_parent is not None:
+                        complementary_msgs.append(parent.msg_from_parent)
+
+                    if hasattr(self, 'merger_model') and self.merger_model:
+                        time_points = np.unique(np.concatenate([msg.x for msg in complementary_msgs]))
+                        # As Lx do not include the node contribution this must be added on
+                        complementary_msgs.append(self.merger_model.node_contribution(parent, time_points))
+
+                        # Removed merger rate must be added back if no msgs from parent (equivalent to root node case)
+                        if parent.msg_from_parent is None:
+                            complementary_msgs.append(Distribution(time_points, self.merger_model.integral_merger_rate(time_points), is_log=True))
+
+                    if len(complementary_msgs):
+                        msg_parent_to_node = NodeInterpolator.multiply(complementary_msgs)
+                        msg_parent_to_node._adjust_grid(rel_tol=self.rel_tol_prune)
+
+                elif parent.marginal_pos_LH is not None:
+                    msg_parent_to_node = parent.marginal_pos_LH
+
+                if msg_parent_to_node is None:
                     x = [parent.numdate, numeric_date()]
                     msg_parent_to_node = NodeInterpolator(x, [1.0, 1.0],min_width=self.min_width)
 
                 # integral message, which delivers to the node the positional information
                 # from the complementary subtree
-                res, res_t = NodeInterpolator.convolve(msg_parent_to_node, node.branch_length_interpolator,
-                                                    max_or_integral='integral',
-                                                    inverse_time=False,
-                                                    n_grid_points = self.node_grid_points,
-                                                    n_integral=self.n_integral,
-                                                    rel_tol=self.rel_tol_refine)
+                if self.use_fft:
+                    res, res_t = NodeInterpolator.convolve_fft(msg_parent_to_node, node.branch_length_interpolator,
+                                        fft_grid_size = self.fft_grid_size, inverse_time=False), None
+                else:
+                    res, res_t = NodeInterpolator.convolve(msg_parent_to_node, node.branch_length_interpolator,
+                                        max_or_integral='integral',
+                                        inverse_time=False,
+                                        n_grid_points = self.node_grid_points,
+                                        n_integral=self.n_integral,
+                                        rel_tol=self.rel_tol_refine)
 
                 node.msg_from_parent = res
                 if node.marginal_pos_Lx is None:
@@ -626,7 +744,7 @@ class ClockTree(TreeAnc):
                         plt.plot(msg_parent_to_node.x,msg_parent_to_node.y-msg_parent_to_node.peak_val, '-o')
                         plt.ylim(0,100)
                         plt.xlim(-0.05, 0.05)
-                        import ipdb; ipdb.set_trace()
+                        #import ipdb; ipdb.set_trace()
 
             # assign positions of nodes and branch length only when desired
             # since marginal reconstruction can result in negative branch length
@@ -650,8 +768,6 @@ class ClockTree(TreeAnc):
         if not self.debug:
             _cleanup()
 
-        return
-
 
     def convert_dates(self):
         '''
@@ -728,7 +844,7 @@ class ClockTree(TreeAnc):
 
         self.logger("###ClockTree.calc_rate_susceptibility: run with upper bound of rate estimate", 1)
         self.make_time_tree(**params)
-        self.logger("###ClockTree.calc_rate_susceptibility: rate: %f, LH:%f"%(upper_rate, self.tree.positional_joint_LH), 2)
+        self.logger("###ClockTree.calc_rate_susceptibility: rate: %f, LH:%f"%(upper_rate, self.tree.positional_LH), 2)
         for n in self.tree.find_clades():
             n.numdate_rate_variation = [(upper_rate, n.numdate)]
             if n.up:
@@ -736,7 +852,7 @@ class ClockTree(TreeAnc):
 
         self.logger("###ClockTree.calc_rate_susceptibility: run with lower bound of rate estimate", 1)
         self.make_time_tree(**params)
-        self.logger("###ClockTree.calc_rate_susceptibility: rate: %f, LH:%f"%(lower_rate, self.tree.positional_joint_LH), 2)
+        self.logger("###ClockTree.calc_rate_susceptibility: rate: %f, LH:%f"%(lower_rate, self.tree.positional_LH), 2)
         for n in self.tree.find_clades():
             n.numdate_rate_variation.append((lower_rate, n.numdate))
             if n.up:
@@ -744,7 +860,7 @@ class ClockTree(TreeAnc):
 
         self.logger("###ClockTree.calc_rate_susceptibility: run with central rate estimate", 1)
         self.make_time_tree(**params)
-        self.logger("###ClockTree.calc_rate_susceptibility: rate: %f, LH:%f"%(current_rate, self.tree.positional_joint_LH), 2)
+        self.logger("###ClockTree.calc_rate_susceptibility: rate: %f, LH:%f"%(current_rate, self.tree.positional_LH), 2)
         for n in self.tree.find_clades():
             n.numdate_rate_variation.append((current_rate, n.numdate))
             n.numdate_rate_variation.sort(key=lambda x:x[1]) # sort estimates for different rates by numdate
@@ -874,7 +990,7 @@ class ClockTree(TreeAnc):
             pidx = np.argmin(node.marginal_pos_LH.y)
             pval = np.min(node.marginal_pos_LH.y)
 
-            # check if the distribution as at least 3 points and that the peak is not either of the two 
+            # check if the distribution as at least 3 points and that the peak is not either of the two
             # end points. Otherwise, interpolation objects can be initialized.
             if node.marginal_pos_LH.y.shape[0]<3 or pidx==0 or pidx==node.marginal_pos_LH.y.shape[0]-1:
                 value_str = "values: " + ','.join([str(x) for x in node.marginal_pos_LH.y])
