@@ -1,9 +1,8 @@
-from __future__ import division, print_function, absolute_import
 import sys
 from os.path import isfile
 from collections import defaultdict
 import numpy as np
-from Bio import SeqRecord, Seq, AlignIO, SeqIO
+from Bio import AlignIO, SeqIO
 from . import config as ttconf
 from . import MissingDataError
 from .seq_utils import seq2array, guess_alphabet, alphabets
@@ -91,7 +90,7 @@ class SequenceData(object):
         self._ref = None
         self.likely_alphabet = None
         self.compressed_to_full_sequence_map = None
-        self.multiplicity = None
+        self._multiplicity = None
         self.is_sparse = None
         self.convert_upper = convert_upper
         self.compress = compress
@@ -161,9 +160,20 @@ class SequenceData(object):
 
         if type(in_aln) is MultipleSeqAlignment:
             # check whether the alignment is consistent with a nucleotide alignment.
-            self._aln = {s.name: seq2array(s, convert_upper=self.convert_upper,
-                                           fill_overhangs=self.fill_overhangs, ambiguous=self.ambiguous)
-                         for s in in_aln}
+            self._aln = {}
+            for s in in_aln:
+                if s.id==s.name:
+                    tmp_name  = s.id
+                elif '<unknown' in s.id:  # use s.name if id is BioPython default (previous behavior)
+                    tmp_name = s.name
+                elif '<unknown' in s.name:  # use s.id if s.name is BioPython default (change relative to previous, but what we want)
+                    tmp_name = s.id
+                else:
+                    tmp_name = s.name  # otherwise use s.name (previous behavior)
+
+                self._aln[tmp_name] = seq2array(s, convert_upper=self.convert_upper,
+                                               fill_overhangs=self.fill_overhangs, ambiguous=self.ambiguous)
+
             self.check_alphabet(list(self._aln.values()))
             self.is_sparse = False
             self.logger("SequenceData: loaded alignment.",1)
@@ -260,8 +270,13 @@ class SequenceData(object):
             self._ref = seq2array(in_ref, fill_overhangs=False, word_length=self.word_length)
             self.full_length = self._ref.shape[0]
             self.compressed_to_full_sequence_map = None
-            self.multiplicity = None
+            self._multiplicity = None
 
+    def multiplicity(self, mask=None):
+        if mask is None:
+            return self._multiplicity
+        else:
+            return self._multiplicity*mask
 
     def check_alphabet(self, seqs):
         self.likely_alphabet = guess_alphabet(seqs)
@@ -299,7 +314,7 @@ class SequenceData(object):
         """
 
         if not self.compress: #
-            self.multiplicity = np.ones(self.full_length, dtype=float)
+            self._multiplicity = np.ones(self.full_length, dtype=float)
             self.full_to_compressed_sequence_map = np.arange(self.full_length)
             self.compressed_to_full_sequence_map = {p:np.array([p]) for p in np.arange(self.full_length)}
             self._compressed_length = self._full_length
@@ -332,7 +347,7 @@ class SequenceData(object):
         for pi in variable_positions:
             if self.is_sparse:
                 pattern = np.array([self.aln[k][pi] if pi in self.aln[k] else self.ref[pi]
-                           for k in self.sequence_names], dtype='S')
+                           for k in self.sequence_names])
             else:
                 # pylint: disable=unsubscriptable-object
                 pattern = np.copy(aln_transpose[pi])
@@ -391,9 +406,9 @@ class SequenceData(object):
 
 
         # count how many times each column is repeated in the real alignment
-        self.multiplicity = np.zeros(len(alignment_patterns))
+        self._multiplicity = np.zeros(len(alignment_patterns))
         for p, pos in alignment_patterns.values():
-            self.multiplicity[p]=len(pos)
+            self._multiplicity[p]=len(pos)
 
         # create the compressed alignment as a dictionary linking names to sequences
         tmp_compressed_alignment = np.array(compressed_aln_transpose).T
@@ -410,7 +425,7 @@ class SequenceData(object):
             self.compressed_to_full_sequence_map[val[0]]=np.array(val[1], dtype=int)
 
         self.logger("SequenceData: constructed compressed alignment...", 1)
-        self._compressed_length = len(self.multiplicity)
+        self._compressed_length = len(self._multiplicity)
         return ttconf.SUCCESS
 
 
@@ -488,7 +503,7 @@ class SequenceData(object):
         else:
             return tmp_seq
 
-    def differences(self, seq1, seq2, seq1_compressed=True, seq2_compressed=True):
+    def differences(self, seq1, seq2, seq1_compressed=True, seq2_compressed=True, mask=None):
         diffs = []
         if self.is_sparse:
             if seq1_compressed: seq1 = self.compressed_to_sparse_sequence(seq1)
@@ -503,7 +518,11 @@ class SequenceData(object):
         else:
             if seq1_compressed: seq1 = self.compressed_to_full_sequence(seq1)
             if seq2_compressed: seq2 = self.compressed_to_full_sequence(seq2)
-            diff_pos = np.where(seq1 != seq2)[0]
+            if mask is None:
+                diff_pos = np.where(seq1 != seq2)[0]
+            else:
+                diff_pos = np.where((seq1 != seq2)&(mask>0))[0]
+
             for pos in diff_pos:
                 diffs.append((seq1[pos], pos, seq2[pos]))
 
