@@ -1,14 +1,12 @@
-from __future__ import print_function, division, absolute_import
 import time, sys
 import gc
-from collections import defaultdict
 import numpy as np
 from Bio import Phylo
 from Bio.Phylo.BaseTree import Clade
 from Bio import AlignIO
 from . import config as ttconf
 from . import MissingDataError,UnknownMethodError
-from .seq_utils import seq2prof, seq2array, prof2seq, normalize_profile, extend_profile
+from .seq_utils import seq2prof, prof2seq, normalize_profile, extend_profile
 from .gtr import GTR
 from .gtr_site_specific import GTR_site_specific
 from .sequence_data import SequenceData
@@ -34,11 +32,11 @@ def mutations(node):
     if node.up is None:
         return []
     elif (not node.tt.reconstructed_tip_sequences) and node.name in node.tt.data.aln:
-        return node.tt.data.differences(node.up.cseq, node.tt.data.aln[node.name], seq2_compressed=False)
+        return node.tt.data.differences(node.up.cseq, node.tt.data.aln[node.name], seq2_compressed=False, mask=node.mask)
     elif node.is_terminal() and (node.name not in node.tt.data.aln):
         return []
     else:
-        return node.tt.data.differences(node.up.cseq, node.cseq)
+        return node.tt.data.differences(node.up.cseq, node.cseq, mask=node.mask)
 
 string_types = [str] if sys.version_info[0]==3 else [str, unicode]
 Clade.sequence = property(lambda x: x.tt.sequence(x, as_string=False))
@@ -56,8 +54,8 @@ class TreeAnc(object):
     def __init__(self, tree=None, aln=None, gtr=None, fill_overhangs=True,
                 ref=None, verbose = ttconf.VERBOSE, ignore_gaps=True,
                 convert_upper=True, seq_multiplicity=None, log=None,
-                 compress=True, seq_len=None,
-                **kwargs):
+                compress=True, seq_len=None, ignore_missing_alns=False,
+                keep_node_order=False, **kwargs):
         """
         TreeAnc constructor. It prepares the tree, attaches sequences to the leaf nodes,
         and sets some configuration parameters.
@@ -78,22 +76,22 @@ class TreeAnc(object):
            GTR model object. If string passed, it is interpreted as the type of
            the GTR model. A new GTR instance will be created for this type.
 
-        fill_overhangs : bool
+        fill_overhangs : bool, default True
            In some cases, the missing data on both ends of the alignment is
            filled with the gap sign('-'). If set to True, the end-gaps are converted to "unknown"
            characters ('N' for nucleotides, 'X' for aminoacids). Otherwise, the alignment is treated as-is
 
         ref : None, optional
-            Reference sequence used in VCF mode
+           Reference sequence used in VCF mode
 
-        verbose : int
+        verbose : int, default 3
            Verbosity level as number from 0 (lowest) to 10 (highest).
 
-        ignore_gaps : bool
+        ignore_gaps : bool, default True
            Ignore gaps in branch length calculations
 
-        convert_upper : bool, optional
-            Description
+        convert_upper : bool, default True
+           Convert all sequences to upper case
 
         seq_multiplicity : dict
            If individual nodes in the tree correspond to multiple sampled sequences
@@ -101,7 +99,7 @@ class TreeAnc(object):
            specified as a dictionary. This currently only affects rooting and
            can be used to weigh individual tips by abundance or important during root search.
 
-        compress : bool, optional
+        compress : bool, default True
             reduce identical alignment columns to one (not useful when
             inferring site specific GTR models).
 
@@ -109,6 +107,8 @@ class TreeAnc(object):
             length of the sequence. this is inferred from the input alignment or the reference
             sequence in most cases but can be specified for other applications.
 
+        ignore_missing_alns : bool, default False
+
         **kwargs
            Keyword arguments to construct the GTR model
 
@@ -139,6 +139,8 @@ class TreeAnc(object):
         self.ignore_gaps = ignore_gaps
         self.reconstructed_tip_sequences = False
         self.sequence_reconstruction = None
+        self.ignore_missing_alns = ignore_missing_alns
+        self.keep_node_order = keep_node_order
 
         self._tree = None
         self.tree = tree
@@ -335,7 +337,7 @@ class TreeAnc(object):
         Returns
         -------
         float
-            inverse of the uncompressed sequene length - length scale for short branches
+            inverse of the uncompressed sequence length - length scale for short branches
         """
         return 1.0/self.data.full_length if self.data.full_length else np.nan
 
@@ -374,9 +376,9 @@ class TreeAnc(object):
         for l in self.tree.find_clades():
             if hasattr(l, 'branch_state'): del l.branch_state
             if l.name not in self.data.compressed_alignment and l.is_terminal():
-                self.logger("***WARNING: TreeAnc._attach_sequences_to_nodes: NO SEQUENCE FOR LEAF: %s" % l.name, 0, warn=True)
+                self.logger("***WARNING: TreeAnc._check_alignment_tree_gtr_consistency: NO SEQUENCE FOR LEAF: '%s'" % l.name, 0, warn=True)
                 failed_leaves += 1
-                if failed_leaves > self.tree.count_terminals()/3:
+                if not self.ignore_missing_alns and failed_leaves > self.tree.count_terminals()/3:
                     raise MissingDataError("TreeAnc._check_alignment_tree_gtr_consistency: At least 30\\% terminal nodes cannot be assigned a sequence!\n"
                                            "Are you sure the alignment belongs to the tree?")
             else: # could not assign sequence for internal node - is OK
@@ -400,8 +402,10 @@ class TreeAnc(object):
         """
         self.sequence_reconstruction = False
         self.tree.root.branch_length = 0.001
+        self.tree.root.mask = None
         self.tree.root.mutation_length = self.tree.root.branch_length
-        self.tree.ladderize()
+        if not self.keep_node_order:
+            self.tree.ladderize()
         self._prepare_nodes()
         self._leaves_lookup = {node.name:node for node in self.tree.get_terminals()}
 
@@ -429,11 +433,14 @@ class TreeAnc(object):
                 c.up = clade
                 c.tt = self
 
+
         for clade in self.tree.find_clades(order='postorder'): # children first
             if clade.is_terminal():
                 clade.bad_branch = clade.bad_branch if hasattr(clade, 'bad_branch') else False
             else:
                 clade.bad_branch = all([c.bad_branch for c in clade])
+            if not hasattr(clade, "mask"):
+                clade.mask = None
 
         self._calc_dist2root()
         self._internal_node_count = max(internal_node_count, self._internal_node_count)
@@ -682,7 +689,7 @@ class TreeAnc(object):
          log_lh : float
             The tree likelihood given the sequences
         """
-        log_lh = np.zeros(self.data.multiplicity.shape[0])
+        log_lh = np.zeros(self.data.multiplicity().shape[0])
         for node in self.tree.find_clades(order='postorder'):
 
             if node.up is None: #  root node
@@ -781,7 +788,7 @@ class TreeAnc(object):
         marginal_LH_prefactor = self.tree.root.marginal_subtree_LH_prefactor + pre
 
         self.tree.sequence_LH = marginal_LH_prefactor
-        self.tree.total_sequence_LH = (self.tree.sequence_LH*self.data.multiplicity).sum()
+        self.tree.total_sequence_LH = (self.tree.sequence_LH*self.data.multiplicity()).sum()
         self.tree.sequence_marginal_LH = self.tree.total_sequence_LH
         if assign_sequence:
             seq, prof_vals, idxs = prof2seq(self.tree.root.marginal_profile,
@@ -811,8 +818,13 @@ class TreeAnc(object):
             tmp_log_subtree_LH = np.zeros((L,n_states), dtype=float)
             node.marginal_subtree_LH_prefactor = np.zeros(L, dtype=float)
             for ch in node.clades:
-                ch.marginal_log_Lx = self.gtr.propagate_profile(ch.marginal_subtree_LH,
-                    self._branch_length_to_gtr(ch), return_log=True) # raw prob to transfer prob up
+                if ch.mask is None:
+                    ch.marginal_log_Lx = self.gtr.propagate_profile(ch.marginal_subtree_LH,
+                        self._branch_length_to_gtr(ch), return_log=True) # raw prob to transfer prob up
+                else:
+                    ch.marginal_log_Lx = (self.gtr.propagate_profile(ch.marginal_subtree_LH,
+                                self._branch_length_to_gtr(ch), return_log=True).T*ch.mask).T # raw prob to transfer prob up
+
                 tmp_log_subtree_LH += ch.marginal_log_Lx
                 node.marginal_subtree_LH_prefactor += ch.marginal_subtree_LH_prefactor
 
@@ -839,7 +851,13 @@ class TreeAnc(object):
 
             tmp_msg_from_parent = self.gtr.evolve(node.marginal_outgroup_LH,
                                                  self._branch_length_to_gtr(node), return_log=False)
-            node.marginal_profile, pre = normalize_profile(node.marginal_subtree_LH * tmp_msg_from_parent, return_offset=False)
+
+            if node.mask is None:
+                node.marginal_profile, pre = normalize_profile(node.marginal_subtree_LH * tmp_msg_from_parent, return_offset=False)
+            else:
+                node.marginal_profile, pre = normalize_profile(node.marginal_subtree_LH * (node.mask*tmp_msg_from_parent.T + (1.0-node.mask)).T,
+                                                    return_offset=False)
+
             # choose sequence based maximal marginal LH.
             if assign_sequence:
                 seq, prof_vals, idxs = prof2seq(node.marginal_profile, self.gtr,
@@ -906,20 +924,34 @@ class TreeAnc(object):
                 # this is prod_ch L_x(i)
                 msg_from_children = np.sum(np.stack([c.joint_Lx for c in node.clades], axis=0), axis=0)
 
+                if not debug:
+                    # Now that we have calculated the current node's likelihood
+                    # from its children, clean up likelihood matrices attached
+                    # to children to save memory.
+                    for c in node.clades:
+                        del c.joint_Lx
+
             # for every possible state of the parent node,
             # get the best state of the current node
             # and compute the likelihood of this state
             # preallocate storage
-            node.joint_Lx = np.zeros((L, n_states))             # likelihood array
-            node.joint_Cx = np.zeros((L, n_states), dtype=int)  # max LH indices
+            node.joint_Lx = np.zeros((L, n_states)) # likelihood array
+            node.joint_Cx = np.zeros((L, n_states), dtype=np.uint16)  # max LH indices
             for char_i, char in enumerate(self.gtr.alphabet):
                 # Pij(i) * L_ch(i) for given parent state j
-                msg_to_parent = (log_transitions[:,char_i].T + msg_from_children)
-                # For this parent state, choose the best state of the current node:
+                # if the node has a mask, P_ij is uniformly 1 at masked positions as no info is propagated
+                if node.mask is None:
+                    msg_to_parent = (log_transitions[:,char_i].T + msg_from_children)
+                else:
+                    msg_to_parent = ((log_transitions[:,char_i]*np.repeat([node.mask], self.gtr.n_states, axis=0).T) + msg_from_children)
+
+                # For this parent state, choose the best state of the current node
                 node.joint_Cx[:, char_i] = msg_to_parent.argmax(axis=1)
-                # compute the likelihood of the best state of the current node
-                # given the state of the parent (char_i)
+                # and compute the likelihood of the best state of the current node
+                # given the state of the parent (char_i) -- at masked position, there is no contribution
                 node.joint_Lx[:, char_i] = msg_to_parent.max(axis=1)
+                if node.mask is not None:
+                    node.joint_Lx[:, char_i] *= node.mask
 
         # root node profile = likelihood of the total tree
         msg_from_children = np.sum(np.stack([c.joint_Lx for c in self.tree.root], axis = 0), axis=0)
@@ -939,7 +971,7 @@ class TreeAnc(object):
 
         # compute the likelihood of the most probable root sequence
         self.tree.sequence_LH = np.choose(idxs, self.tree.root.joint_Lx.T)
-        self.tree.sequence_joint_LH = (self.tree.sequence_LH*self.data.multiplicity).sum()
+        self.tree.sequence_joint_LH = (self.tree.sequence_LH*self.data.multiplicity()).sum()
         self.tree.root._cseq = seq
         self.tree.root.seq_idx = idxs
 
@@ -973,7 +1005,10 @@ class TreeAnc(object):
         # do clean-up
         if not debug:
             for node in self.tree.find_clades(order='preorder'):
-                del node.joint_Lx
+                # Check for the likelihood matrix, since we might have cleaned
+                # it up earlier.
+                if hasattr(node, "joint_Lx"):
+                    del node.joint_Lx
                 del node.joint_Cx
                 if hasattr(node, 'seq_idx'):
                     del node.seq_idx
@@ -1061,7 +1096,7 @@ class TreeAnc(object):
         """
         seq_pairs, multiplicity = self.gtr.state_pair(
                                        node.up.cseq, node.cseq,
-                                       pattern_multiplicity = self.data.multiplicity,
+                                       pattern_multiplicity = self.data.multiplicity(mask=node.mask),
                                        ignore_gaps = self.ignore_gaps)
         node.branch_state = {'pair':seq_pairs, 'multiplicity':multiplicity}
 
@@ -1181,7 +1216,7 @@ class TreeAnc(object):
             return self.one_mutation
         else:
             pp, pc = self.marginal_branch_profile(node)
-            return self.gtr.optimal_t_compressed((pp, pc), self.data.multiplicity, profiles=True, tol=tol)
+            return self.gtr.optimal_t_compressed((pp, pc), self.data.multiplicity(mask=node.mask), profiles=True, tol=tol)
 
 
     def optimize_tree_marginal(self, max_iter=10, infer_gtr=False, pc=1.0, damping=0.75,
@@ -1209,7 +1244,10 @@ class TreeAnc(object):
                     prof_c = n1.marginal_subtree_LH
                     prof_p = normalize_profile(n2.marginal_subtree_LH*self.tree.root.marginal_outgroup_LH)[0]
 
-                    new_bl = self.gtr.optimal_t_compressed((prof_p, prof_c), self.data.multiplicity, profiles=True, tol=tol)
+                    if n1.mask is None or n2.mask is None:
+                        new_bl = self.gtr.optimal_t_compressed((prof_p, prof_c), self.data.multiplicity(), profiles=True, tol=tol)
+                    else:
+                        new_bl = self.gtr.optimal_t_compressed((prof_p, prof_c), self.data.multiplicity(mask=n1.mask*n2.mask), profiles=True, tol=tol)
                     update_val = new_bl*(1-damping**(i+1)) +  total_bl*damping**(i+1)
 
                     n1.branch_length = update_val*bl_ratio
@@ -1247,9 +1285,8 @@ class TreeAnc(object):
         self.logger("Deprecation warning: 'optimize_seq_and_branch_len' will be removed and replaced by 'optimize_tree'!", 1, warn=True)
         self.optimize_tree(*args,**kwargs)
 
-    def optimize_tree(self,prune_short=True,
-                      marginal_sequences=False, branch_length_mode='joint',
-                      max_iter=5, infer_gtr=False, pc=1.0, **kwargs):
+    def optimize_tree(self,prune_short=True, marginal_sequences=False, branch_length_mode='joint',
+                      max_iter=5, infer_gtr=False, pc=1.0, method_anc='probabilistic', **kwargs):
         """
         Iteratively set branch lengths and reconstruct ancestral sequences until
         the values of either former or latter do not change. The algorithm assumes
@@ -1284,6 +1321,9 @@ class TreeAnc(object):
          infer_gtr : bool
             Infer a GTR model from the observed substitutions.
 
+         method_anc: str
+            Which method should be used to reconstruct ancestral sequences.
+            Supported values are "parsimony", "fitch", "probabilistic" and "ml"
         """
         if branch_length_mode=='marginal':
             self.optimize_tree_marginal(max_iter=max_iter, infer_gtr=infer_gtr, pc=pc, **kwargs)
@@ -1291,7 +1331,7 @@ class TreeAnc(object):
                 self.prune_short_branches()
             return ttconf.SUCCESS
         elif branch_length_mode=='input':
-            N_diff = self.reconstruct_anc(method='probabilistic', infer_gtr=infer_gtr, pc=pc,
+            N_diff = self.reconstruct_anc(method=method_anc, infer_gtr=infer_gtr, pc=pc,
                                           marginal=marginal_sequences, **kwargs)
             if prune_short:
                 self.prune_short_branches()
@@ -1300,7 +1340,7 @@ class TreeAnc(object):
             raise UnknownMethodError("TreeAnc.optimize_tree: `branch_length_mode` should be in ['marginal', 'joint', 'input']")
 
         self.logger("TreeAnc.optimize_tree: sequences...", 1)
-        N_diff = self.reconstruct_anc(method='probabilistic', infer_gtr=infer_gtr, pc=pc,
+        N_diff = self.reconstruct_anc(method=method_anc, infer_gtr=infer_gtr, pc=pc,
                                       marginal=marginal_sequences, **kwargs)
         self.optimize_branch_len(verbose=0, store_old=False, mode=branch_length_mode)
 
@@ -1309,7 +1349,7 @@ class TreeAnc(object):
             n += 1
             if prune_short:
                 self.prune_short_branches()
-            N_diff = self.reconstruct_anc(method='probabilistic', infer_gtr=False,
+            N_diff = self.reconstruct_anc(method=method_anc, infer_gtr=False,
                                           marginal=marginal_sequences, **kwargs)
 
             self.logger("TreeAnc.optimize_tree: Iteration %d."
@@ -1319,7 +1359,7 @@ class TreeAnc(object):
                 break
             self.optimize_branch_lengths_joint(store_old=False)
 
-        self.tree.unconstrained_sequence_LH = (self.tree.sequence_LH*self.data.multiplicity).sum()
+        self.tree.unconstrained_sequence_LH = (self.tree.sequence_LH*self.data.multiplicity()).sum()
         self._prepare_nodes() # fix dist2root and up-links after reconstruction
         self.logger("TreeAnc.optimize_tree: Unconstrained sequence LH:%f" % self.tree.unconstrained_sequence_LH , 2)
         return ttconf.SUCCESS
@@ -1338,7 +1378,7 @@ class TreeAnc(object):
             # probability of the two seqs separated by zero time is not zero
             if  ((node.branch_length<0.1*self.one_mutation) and
                  (self.gtr.prob_t(node.up._cseq, node._cseq, 0.0,
-                                  pattern_multiplicity=self.data.multiplicity) > 0.1)):
+                                  pattern_multiplicity=self.data.multiplicity(mask=node.mask)) > 0.1)):
                 # re-assign the node children directly to its parent
                 node.up.clades = [k for k in node.up.clades if k != node] + node.clades
                 for clade in node.clades:
@@ -1409,9 +1449,9 @@ class TreeAnc(object):
             for c in node:
                 if marginal:
                     mut_stack = np.transpose(self.get_branch_mutation_matrix(c, full_sequence=False), (1,2,0))
-                    T_ia += 0.5*self._branch_length_to_gtr(c) * mut_stack.sum(axis=0) * self.data.multiplicity
-                    T_ia += 0.5*self._branch_length_to_gtr(c) * mut_stack.sum(axis=1) * self.data.multiplicity
-                    n_ija += mut_stack * self.data.multiplicity
+                    T_ia += 0.5*self._branch_length_to_gtr(c) * mut_stack.sum(axis=0) * self.data.multiplicity(mask=c.mask)
+                    T_ia += 0.5*self._branch_length_to_gtr(c) * mut_stack.sum(axis=1) * self.data.multiplicity(mask=c.mask)
+                    n_ija += mut_stack * self.data.multiplicity(mask=c.mask)
                 else:
                     for a,pos, d in c.mutations:
                         try:
@@ -1428,7 +1468,7 @@ class TreeAnc(object):
                         cseq = c.cseq
                         if cseq is not None:
                             ind = cseq==nuc
-                            T_ia[i,ind] += self._branch_length_to_gtr(c)*self.data.multiplicity[ind]
+                            T_ia[i,ind] += self._branch_length_to_gtr(c)*self.data.multiplicity(mask=c.mask)[ind]
 
         self.logger("TreeAnc.infer_gtr: counting mutations...done", 3)
 
@@ -1441,7 +1481,7 @@ class TreeAnc(object):
                                 root_state=root_state, logger=self.logger,
                                 alphabet=self.gtr.alphabet, prof_map=self.gtr.profile_map)
         else:
-            root_state = np.array([np.sum((self.tree.root.cseq==nuc)*self.data.multiplicity)
+            root_state = np.array([np.sum((self.tree.root.cseq==nuc)*self.data.multiplicity(mask=self.tree.root.mask))
                                    for nuc in self.gtr.alphabet])
             n_ij = n_ija.sum(axis=-1)
             self._gtr = GTR.infer(n_ij, T_ia.sum(axis=-1), root_state, fixed_pi=fixed_pi, pc=pc,
