@@ -1,4 +1,3 @@
-from __future__ import print_function, division, absolute_import
 import numpy as np
 from scipy import optimize as sciopt
 from Bio import Phylo
@@ -11,6 +10,24 @@ rerooting_mechanisms = ["min_dev", "best", "least-squares"]
 deprecated_rerooting_mechanisms = {"residual":"least-squares", "res":"least-squares",
                                    "min_dev_ML": "min_dev", "ML":"least-squares"}
 
+
+def reduce_time_marginal_argument(input_time_marginal):
+    '''
+    This function maps deprecated arguments/terms for the timetree inference mode
+    to recommended terms.
+    '''
+    if input_time_marginal in [False, 'false', 'never']:
+        return 'never'
+    elif input_time_marginal in [True, 'always', 'true']:
+        return 'always'
+    elif input_time_marginal in ['only-final', 'assign']:
+        return 'only-final'
+    elif input_time_marginal == 'confidence-only':
+        return input_time_marginal
+    else:
+        raise ValueError(f"'{input_time_marginal}' is not a known time marginal argument")
+
+
 class TreeTime(ClockTree):
     """
     TreeTime is a wrapper class to ClockTree that adds additional functionality
@@ -36,8 +53,9 @@ class TreeTime(ClockTree):
 
     def run(self, root=None, infer_gtr=True, relaxed_clock=None, n_iqd = None,
             resolve_polytomies=True, max_iter=0, Tc=None, fixed_clock_rate=None,
-            time_marginal=False, sequence_marginal=False, branch_length_mode='auto',
-            vary_rate=False, use_covariation=False, tracelog_file=None, **kwargs):
+            time_marginal='never', sequence_marginal=False, branch_length_mode='auto',
+            vary_rate=False, use_covariation=False, tracelog_file=None, 
+            method_anc = 'probabilistic', **kwargs):
 
         """
         Run TreeTime reconstruction. Based on the input parameters, it divides
@@ -85,8 +103,11 @@ class TreeTime(ClockTree):
         fixed_clock_rate : float
            Fixed clock rate to be used. If None, infer clock rate from the molecular clock.
 
-        time_marginal : bool
-           If True, perform a final round of marginal reconstruction of the node's positions.
+        time_marginal : bool, str
+           If False perform joint reconstruction of the divergence times, if True use marginal
+           reconstruction of the divergence times, if 'only_final' (or 'assign') apply the marginal reconstruction
+           only to the last optimization round, if "confidence-only" perform additional round using marginal
+           reconstruction for calculation of confidence intervals but do not update times.
 
         sequence_marginal : bool, optional
             use marginal reconstruction for ancestral sequences
@@ -109,6 +130,11 @@ class TreeTime(ClockTree):
             regression ignoring phylogenetic covaration between nodes. If vary_rate is True,
             use_covariation is true by default
 
+        method_anc: str, optional
+            Which method should be used to reconstruct ancestral sequences. 
+            Supported values are "parsimony", "fitch", "probabilistic" and "ml". 
+            Default is "probabilistic"
+
         **kwargs
            Keyword arguments needed by the downstream functions
 
@@ -120,7 +146,6 @@ class TreeTime(ClockTree):
 
 
         """
-
         # register the specified covaration mode
         self.use_covariation = use_covariation or (vary_rate and (not type(vary_rate)==float))
 
@@ -128,16 +153,17 @@ class TreeTime(ClockTree):
             raise MissingDataError("TreeTime.run: ERROR, alignment or tree are missing")
         if self.aln is None:
             branch_length_mode='input'
-
         self._set_branch_length_mode(branch_length_mode)
 
         # determine how to reconstruct and sample sequences
         seq_kwargs = {"marginal_sequences":sequence_marginal or (self.branch_length_mode=='marginal'),
                       "branch_length_mode": self.branch_length_mode,
                       "sample_from_profile":"root",
+                      "prune_short":kwargs.get("prune_short", True),
                       "reconstruct_tip_states":kwargs.get("reconstruct_tip_states", False)}
-
-        tt_kwargs = {'clock_rate':fixed_clock_rate, 'time_marginal':False}
+        time_marginal_method = reduce_time_marginal_argument(time_marginal) ## for backward compatibility
+        tt_kwargs = {'clock_rate':fixed_clock_rate,
+                     'time_marginal':False if time_marginal_method in ['never', 'only-final', 'confidence-only'] else True}
         tt_kwargs.update(kwargs)
 
         seq_LH = 0
@@ -150,10 +176,11 @@ class TreeTime(ClockTree):
         if self.branch_length_mode=='input':
             if self.aln:
                 self.infer_ancestral_sequences(infer_gtr=infer_gtr, marginal=seq_kwargs["marginal_sequences"], **seq_kwargs)
-                self.prune_short_branches()
+                if seq_kwargs["prune_short"]:
+                    self.prune_short_branches()
         else:
             self.optimize_tree(infer_gtr=infer_gtr,
-                               max_iter=1, prune_short=True, **seq_kwargs)
+                               max_iter=1, method_anc = method_anc, **seq_kwargs)
         avg_root_to_tip = np.mean([x.dist2root for x in self.tree.get_terminals()])
 
         # optionally reroot the tree either by oldest, best regression or with a specific leaf
@@ -171,7 +198,7 @@ class TreeTime(ClockTree):
             if self.aln:
                 self.infer_ancestral_sequences(**seq_kwargs)
         else:
-            self.optimize_tree(max_iter=1, prune_short=False,**seq_kwargs)
+            self.optimize_tree(max_iter=1, method_anc = method_anc, **seq_kwargs)
 
         # infer time tree and optionally resolve polytomies
         self.logger("###TreeTime.run: INITIAL ROUND",0)
@@ -179,10 +206,10 @@ class TreeTime(ClockTree):
 
         if self.aln:
             seq_LH = self.tree.sequence_marginal_LH if seq_kwargs['marginal_sequences'] else self.tree.sequence_joint_LH
-        self.LH =[[seq_LH, self.tree.positional_joint_LH, 0]]
+        self.LH =[[seq_LH, self.tree.positional_LH, 0]]
 
         if root is not None and max_iter:
-            new_root = self.reroot(root='least-squares' if root=='clock_filter' else root, clock_rate=fixed_clock_rate)
+            self.reroot(root='least-squares' if root=='clock_filter' else root, clock_rate=fixed_clock_rate)
             self.logger("###TreeTime.run: rerunning timetree after rerooting",0)
             self.make_time_tree(**tt_kwargs)
 
@@ -212,7 +239,6 @@ class TreeTime(ClockTree):
 
             # estimate a relaxed molecular clock
             if relaxed_clock:
-                print("relaxed_clock", relaxed_clock)
                 self.relaxed_clock(**relaxed_clock)
                 need_new_time_tree = True
 
@@ -223,7 +249,7 @@ class TreeTime(ClockTree):
                 if n_resolved:
                     self.prepare_tree()
                     if self.branch_length_mode!='input': # otherwise reoptimize branch length while preserving branches without mutations
-                        self.optimize_tree(prune_short=False, max_iter=0, **seq_kwargs)
+                        self.optimize_tree(max_iter=0, method_anc = method_anc,**seq_kwargs)
 
                     need_new_time_tree = True
 
@@ -240,7 +266,7 @@ class TreeTime(ClockTree):
 
             if self.aln:
                 seq_LH = self.tree.sequence_marginal_LH if seq_kwargs['marginal_sequences'] else self.tree.sequence_joint_LH
-            self.LH.append([seq_LH, self.tree.positional_joint_LH, self.tree.coalescent_joint_LH])
+            self.LH.append([seq_LH, self.tree.positional_LH, self.tree.coalescent_joint_LH])
 
             # Update the trace log
             self.trace_run.append(self.tracelog_run(niter=niter+1, ndiff=ndiff, n_resolved=n_resolved,
@@ -265,14 +291,15 @@ class TreeTime(ClockTree):
 
         # if marginal reconstruction requested, make one more round with marginal=True
         # this will set marginal_pos_LH, which to be used as error bar estimations
-        if time_marginal:
+        if time_marginal_method in ['only-final', 'confidence-only']:
             self.logger("###TreeTime.run: FINAL ROUND - confidence estimation via marginal reconstruction", 0)
-            tt_kwargs['time_marginal']=time_marginal
-            self.make_time_tree(**tt_kwargs)
+            tt_kwargs['time_marginal'] = True
+            assign_dates = time_marginal_method!='confidence-only'
+            self.make_time_tree(**tt_kwargs, assign_dates=assign_dates)
 
             self.trace_run.append(self.tracelog_run(niter=niter+1, ndiff=0, n_resolved=0,
-                                      time_marginal = tt_kwargs['time_marginal'],
-                                      sequence_marginal = seq_kwargs['marginal_sequences'], Tc=Tc, tracelog=tracelog_file))
+                                      time_marginal=tt_kwargs['time_marginal'],
+                                      sequence_marginal=seq_kwargs['marginal_sequences'], Tc=Tc, tracelog=tracelog_file))
 
         # explicitly print out which branches are bad and whose dates don't correspond to the input dates
         bad_branches =[n for n in self.tree.get_terminals()
@@ -474,6 +501,7 @@ class TreeTime(ClockTree):
             #(Without outgroup_branch_length, gives a trifurcating root, but this will mean
             #mutations may have to occur multiple times.)
             self.tree.root_with_outgroup(new_root, outgroup_branch_length=new_root.branch_length/2)
+            self.tree.root.clades.sort(key = lambda x:x.count_terminals())
             self.get_clock_model(covariation=use_cov, slope = slope)
 
 
@@ -554,7 +582,7 @@ class TreeTime(ClockTree):
         nothing. Otherwise, for each pair of nodes, assess the possible LH increase
         which could be gained by merging the two nodes. The increase in the LH is
         basically the tradeoff between the gain of the LH due to the changing the
-        branch lenghts towards the optimal values and the decrease due to the
+        branch lengths towards the optimal values and the decrease due to the
         introduction of the new branch with zero optimal length.
         """
 
@@ -567,8 +595,8 @@ class TreeTime(ClockTree):
             cost gain if nodes n1, n2 are joined and their parent is placed at time t
             cost gain = (LH loss now) - (LH loss when placed at time t)
             """
-            cg2 = n2.branch_length_interpolator(parent.time_before_present - n2.time_before_present) - n2.branch_length_interpolator(t - n2.time_before_present)
-            cg1 = n1.branch_length_interpolator(parent.time_before_present - n1.time_before_present) - n1.branch_length_interpolator(t - n1.time_before_present)
+            cg2 = n2.branch_length_interpolator._func(parent.time_before_present - n2.time_before_present) - n2.branch_length_interpolator._func(t - n2.time_before_present)
+            cg1 = n1.branch_length_interpolator._func(parent.time_before_present - n1.time_before_present) - n1.branch_length_interpolator._func(t - n1.time_before_present)
             cg_new = - zero_branch_slope * (parent.time_before_present - t) # loss in LH due to the new branch
             return -(cg2+cg1+cg_new)
 
@@ -612,7 +640,16 @@ class TreeTime(ClockTree):
                 # fix positions and branch lengths
                 new_node.time_before_present = new_positions[idxs]
                 new_node.branch_length = clade.time_before_present - new_node.time_before_present
+                self.logger(f"TreeTime._poly.merge_nodes: merging {n1.name} and {n2.name}",4)
                 new_node.clades = [n1,n2]
+                if n1.mask is None or n2.mask is None:
+                    new_node.mask = None
+                    new_node.mcc = None
+                else:
+                    new_node.mask = n1.mask * n2.mask
+                    new_node.mcc = n1.mcc if n1.mcc==n2.mcc else None
+                    self.logger('TreeTime._poly.merge_nodes: assigning mcc to new node ' + new_node.mcc, 4)
+
                 n1.branch_length = new_node.time_before_present - n1.time_before_present
                 n2.branch_length = new_node.time_before_present - n2.time_before_present
 
@@ -632,8 +669,9 @@ class TreeTime(ClockTree):
                     new_node.profile_pair = self.marginal_branch_profile(new_node)
 
                 new_node.branch_length_interpolator = BranchLenInterpolator(new_node, self.gtr,
-                            pattern_multiplicity = self.data.multiplicity, min_width=self.min_width,
-                            one_mutation=self.one_mutation, branch_length_mode=self.branch_length_mode)
+                            pattern_multiplicity = self.data.multiplicity(mask=new_node.mask), min_width=self.min_width,
+                            one_mutation=self.one_mutation, branch_length_mode=self.branch_length_mode,
+                            n_grid_points = self.branch_grid_points)
 
                 clade.clades.remove(n1)
                 clade.clades.remove(n2)
@@ -668,7 +706,7 @@ class TreeTime(ClockTree):
         stretched = [c for c  in clade.clades if c.mutation_length < c.clock_length]
         compressed = [c for c in clade.clades if c not in stretched]
 
-        if len(stretched)==1 and merge_compressed is False:
+        if len(stretched)<2 and merge_compressed is False:
             return 0.0
 
         LH = merge_nodes(stretched, isall=len(stretched)==len(clade.clades))
@@ -691,13 +729,12 @@ class TreeTime(ClockTree):
         """
         try:
             u_lh = self.tree.unconstrained_sequence_LH
+            t_lh = self.tree.positional_LH
             if joint:
                 s_lh = self.tree.sequence_joint_LH
-                t_lh = self.tree.positional_joint_LH
                 c_lh = self.tree.coalescent_joint_LH
             else:
                 s_lh = self.tree.sequence_marginal_LH
-                t_lh = self.tree.positional_marginal_LH
                 c_lh = 0
 
             print ("###  Tree Log-Likelihood  ###\n"
@@ -710,7 +747,7 @@ class TreeTime(ClockTree):
             print("ERROR. Did you run the corresponding inference (joint/marginal)?")
 
 
-    def add_coalescent_model(self, Tc, **kwargs):
+    def add_coalescent_model(self, Tc, n_branches_posterior=False, **kwargs):
         """Add a coalescent model to the tree and optionally optimze
 
         Parameters
@@ -721,8 +758,8 @@ class TreeTime(ClockTree):
         """
         from .merger_models import Coalescent
         self.logger('TreeTime.run: adding coalescent prior with Tc='+str(Tc),1)
-        self.merger_model = Coalescent(self.tree,
-                                       date2dist=self.date2dist, logger=self.logger)
+        self.merger_model = Coalescent(self.tree, date2dist=self.date2dist,
+                                logger=self.logger, n_branches_posterior=n_branches_posterior)
 
         if Tc=='skyline': # restrict skyline model optimization to last iteration
             self.merger_model.optimize_skyline(**kwargs)
@@ -737,7 +774,6 @@ class TreeTime(ClockTree):
                 except:
                     self.logger("setting of coalescent time scale failed", 1, warn=True)
 
-        self.merger_model.attach_to_tree()
 
 
     def relaxed_clock(self, slack=None, coupling=None, **kwargs):
@@ -824,10 +860,10 @@ class TreeTime(ClockTree):
             'Sample'     : niter,
             'ndiff'      : ndiff,
             'n_resolved' : n_resolved,
-            'seq_mode'   : 'marginal' if sequence_marginal else 'joint',
-            'seq_LH'     : self.tree.sequence_marginal_LH if sequence_marginal else self.tree.sequence_joint_LH,
+            'seq_mode'   : ('marginal' if sequence_marginal else 'joint') if self.aln else 'no sequences given',
+            'seq_LH'     : (self.tree.sequence_marginal_LH if sequence_marginal else self.tree.sequence_joint_LH) if self.aln else 0,
             'pos_mode'   : 'marginal' if time_marginal else 'joint',
-            'pos_LH'     : self.tree.positional_marginal_LH if time_marginal else self.tree.positional_joint_LH,
+            'pos_LH'     : self.tree.positional_LH,
             'coal_mode'  : Tc,
             'coal_LH'    : self.tree.coalescent_joint_LH,
         }
@@ -873,7 +909,7 @@ class TreeTime(ClockTree):
             n.branch_length=n.mutation_length
         self.logger("TreeTime._find_best_root: searching for the best root position...",2)
         Treg = self.setup_TreeRegression(covariation=covariation)
-        return Treg.optimal_reroot(force_positive=force_positive, slope=slope)['node']
+        return Treg.optimal_reroot(force_positive=force_positive, slope=slope, keep_node_order=self.keep_node_order)['node']
 
 
 def plot_vs_years(tt, step = None, ax=None, confidence=None, ticks=True, **kwargs):
@@ -982,8 +1018,9 @@ def plot_vs_years(tt, step = None, ax=None, confidence=None, ticks=True, **kwarg
             raise NotReadyError("confidence needs to be either a float (for max posterior region) or a two numbers specifying lower and upper bounds")
 
         for n in tt.tree.find_clades():
-            pos = cfunc(n, confidence)
-            ax.plot(pos-offset, np.ones(len(pos))*n.ypos, lw=3, c=(0.5,0.5,0.5))
+            if not n.bad_branch:
+                pos = cfunc(n, confidence)
+                ax.plot(pos-offset, np.ones(len(pos))*n.ypos, lw=3, c=(0.5,0.5,0.5))
     return fig, ax
 
 def treetime_to_newick(tt, outf):
